// Copyright (c) 2024 Celestino Amoroso (celestino.amoroso@gmail.com).
// All rights reserved.

// func-builtins.go
package expr

import (
	"math"
	"strconv"
)

func isNilFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = args[0] == nil
	return
}

func isIntFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsInteger(args[0])
	return
}

func isFloatFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsFloat(args[0])
	return
}

func isBoolFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsBool(args[0])
	return
}

func isStringFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsString(args[0])
	return
}

func isFractionFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsFract(args[0])
	return
}

func isRationalFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsRational(args[0])
	return
}

func isListFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsList(args[0])
	return
}

func isDictionaryFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result = IsDict(args[0])
	return
}

func boolFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	switch v := args[0].(type) {
	case int64:
		result = (v != 0)
	case *fraction:
		result = v.num != 0
	case float64:
		result = v != 0.0
	case bool:
		result = v
	case string:
		result = len(v) > 0
	default:
		err = errCantConvert(name, v, "bool")
	}
	return
}

func intFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	switch v := args[0].(type) {
	case int64:
		result = v
	case float64:
		result = int64(math.Trunc(v))
	case bool:
		if v {
			result = int64(1)
		} else {
			result = int64(0)
		}
	case string:
		var i int
		if i, err = strconv.Atoi(v); err == nil {
			result = int64(i)
		}
	default:
		err = errCantConvert(name, v, "int")
	}
	return
}

func decFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	switch v := args[0].(type) {
	case int64:
		result = float64(v)
	case float64:
		result = v
	case bool:
		if v {
			result = float64(1)
		} else {
			result = float64(0)
		}
	case string:
		var f float64
		if f, err = strconv.ParseFloat(v, 64); err == nil {
			result = f
		}
	case *fraction:
		result = v.toFloat()
	default:
		err = errCantConvert(name, v, "float")
	}
	return
}

func fractFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	switch v := args[0].(type) {
	case int64:
		var den int64 = 1
		if len(args) > 1 {
			var ok bool
			if den, ok = args[1].(int64); !ok {
				err = errExpectedGot(name, "integer", args[1])
			} else if den == 0 {
				err = errDivisionByZero(name)
			}
		}
		if err == nil {
			result = newFraction(v, den)
		}
	case float64:
		result, err = float64ToFraction(v)
	case bool:
		if v {
			result = newFraction(1, 1)
		} else {
			result = newFraction(0, 1)
		}
	case string:
		result, err = makeGeneratingFraction(v)
	case *fraction:
		result = v
	default:
		err = errCantConvert(name, v, "float")
	}
	return
}

func iteratorFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	return
}

func ImportBuiltinsFuncs(ctx ExprContext) {
	anyParams := []ExprFuncParam{
		newFuncParam(paramValue),
	}

	ctx.RegisterFunc("isNil", newGolangFunctor(isNilFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isInt", newGolangFunctor(isIntFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isFloat", newGolangFunctor(isFloatFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isBool", newGolangFunctor(isBoolFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isString", newGolangFunctor(isStringFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isFract", newGolangFunctor(isFractionFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isRational", newGolangFunctor(isRationalFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isList", newGolangFunctor(isListFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("isDict", newGolangFunctor(isDictionaryFunc), typeBoolean, anyParams)

	ctx.RegisterFunc("bool", newGolangFunctor(boolFunc), typeBoolean, anyParams)
	ctx.RegisterFunc("int", newGolangFunctor(intFunc), typeInt, anyParams)
	ctx.RegisterFunc("dec", newGolangFunctor(decFunc), typeFloat, anyParams)
	ctx.RegisterFunc("fract", newGolangFunctor(fractFunc), typeFraction, []ExprFuncParam{
		newFuncParam(paramValue),
		newFuncParamFlagDef("denominator", pfOptional, 1),
	})
}

func init() {
	registerImport("base", ImportBuiltinsFuncs, "Base expression tools like isNil(), int(), etc.")
}
